// PiTrex Discrete Read/Write Test Software.
// -Reads/Writes 6522 Registers via Raspberry Pi GPIO.
// Kevin Koster, 2019.

// PiTrex Discrete Read & Write Functions

#include <bcm2835.h>
#include <stdio.h>
#include <unistd.h>

/*
  Port Assignments:
  ^^^^^^^^^^^^^^^^^
Descrip.|Bit    |Pin    |
D0      |16     |36     |
D1      |17     |11     |
D2      |18     |12     |
D3      |19     |35     |
D4      |20     |38     |
D5      |21     |40     |
D6      |22     |15     |
D7      |23     |16     |

A0      |7      |26     |
A1      |8      |24     |
A2      |9      |21     |
A3      |10     |19     |
A12     |11     |23     |
A14     |12     |32     |
R/#W    |13     |33     |

PB6     |5      |29     |
RDY(#OE)|4     |18     |
LATCH_EN|25     |22     |
IRQ     |26     |37     |
#HALT	|27	|13		| -Jumper to GND
#D/A    |N/A    |9      | -GND

3V3     |N/A    |1      |
5V      |N/A    |2      |
GND     |N/A    |6      |

Note: Pins counted from 1, bits counted from 0.
*/

// Initialise GPIO for communication with the Vectrex
int vectrexinit ()
{
	//bcm2835_set_debug(1); //Set debugging mode (GPIO not set, but actions reported)
	int result = bcm2835_init();
	if (result)
	{
	 //Set Initial State (ideally set to configured defaults at power-on):
	 //NOTE: This will disable other functions of these GPIO pins
	  //Input/Output Modes:
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL0/4), 0x9240000);	// GPIO Pins 0-6   = INPUT | 7-9   = OUTPUT
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x249);	 // GPIO Pins 10-13 = OUTPUT| 14-19 = INPUT -- Keep Serial Port Active?
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);	 // GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT
	 
	 //bcm2835_gpio_fsel (25,BCM2835_GPIO_FSEL_OUTP); //Force LATCH EN as output
	 
	  //Output Levels:
	 bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low
	 bcm2835_gpio_set (13); // R/#W High
	}
	
	return result;
}

// Read a byte from the Vectrex at the specified address
// --TODO: select only wanted bits of full address.
unsigned char vectrexread (unsigned char address)
{
	unsigned char datain;
	
	//Sets highest 2 bits permanently High as all the addresses have that, ORed with lowest four bits of address. -- Cheated
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | 0x30) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF);	//Set all outputs Low
	
	//Set Data bus to Input mode:
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x249);		// GPIO Pins 10-13 = OUTPUT| 14-19 = INPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8000);		// GPIO Pins 20-24 = INPUT | 25    = OUTPUT | 26-29 = INPUT
	
	bcm2835_gpio_set (13); // R/#W High
	// Set Address Bus
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), shortaddress); // Set Highs
	
	bcm2835_gpio_hen (24); // Detect when pin High
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	bcm2835_gpio_set (25); // LATCH EN High

	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register
	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	
	datain = ( bcm2835_peri_read (bcm2835_gpio + BCM2835_GPLEV0/4) >> 16) & 0xFF; // Read data byte
	bcm2835_gpio_clr (25); // LATCH EN Low
	return datain;
}

// Write a byte to the Vectrex at the specified address
// --TODO: select only wanted bits of full address.
void vectrexwrite (unsigned int address, unsigned char data)
{
	
	//Sets highest 2 bits permanently High as all the addresses have that, ORed with lowest four bits of address. -- Cheated
	//Shifted to position on 32bit bus
	unsigned long shortaddress = ((address & 0x0F) | 0x30) << 7;
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	// Data Output Mode:
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL1/4), 0x9240249);		// GPIO Pins 10-13 = OUTPUT| 14-15 = INPUT | 16-19 = OUTPUT
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPFSEL2/4), 0x8249);			// GPIO Pins 20-23 = OUTPUT | 24 = INPUT | 25    = OUTPUT | 26-29 = INPUT

	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPSET0/4), (data << 16) | shortaddress ); // Data on Data Bus and Address on Address Bus

//	usleep (100);
	
	bcm2835_gpio_ren (24); //Enable detection of RDY (#OE) rising edge.
	bcm2835_gpio_set_eds (24); //Clear Event Detection Register

	usleep(300);
	bcm2835_gpio_set (25); //LATCH EN High

	while (!bcm2835_gpio_eds (24)); //Loop - optimise? - Should use interrupts.
	
	printf("whoo-hoo, RDY went HIGH\n");
	usleep(1000000);
	
	bcm2835_peri_write ((bcm2835_gpio + BCM2835_GPCLR0/4), 0xFFFFFFFF); //LATCH EN Low  R/#W Low
	bcm2835_gpio_set (13); // R/#W High
}

// Wait until interrupt from Vectrex
void vectrexinterrupt (char interrupt)
{
	if (interrupt)
	{
	 bcm2835_gpio_hen (26); // Detect pin High
	 bcm2835_gpio_set_eds (26); //Clear Event Detection Register
	 while (!bcm2835_gpio_eds (26)); //Loop - optimise? - Should use interrupts.
	}
}

int vectrexclose ()
{
	int result = bcm2835_close();
	return result;
}
